///////////////////////////////////////////////
//    FILE: defsim.C
//          procedures and class functions for 
//          simulating mass-spring networks
///////////////////////////////////////////////

#include "tcl.h"
#include "defsim.h"
#include "lib/lib_defs.h"
#include "lib/lib_math.h"
#include "user.h"
#include <stdlib.h>
#include <GL/gl.h>

NETWORK Network;

FVEC dLight;
FVEC plane;
Matrix sm;

///////////////////////////////////////////////////////////
//    PROC: MASS::MASS(int id, float mass, float *posn, float *vel)
//    DOES: constructor for a MASS with initialization
///////////////////////////////////////////////////////////

MASS::MASS(int i, float m, float *p, float *v, float attr)
{
  id = i;
  mass = m;
  for (int n=0; n<3; n++) {
	prevposn[n]=posn[n] = p[n];
	prevvel[n] = vel[n] = v[n];
  }
  type = FREE;
  next = (MASS *) 0;
  slist = (SLIST *) 0;

  attraction = attr;
}

///////////////////////////////////////////////////////////
//    PROC: MASS::print()
//    DOES: prints all info about a mass point
///////////////////////////////////////////////////////////

void MASS::print()
{
  printf("id=%-4d m=%6.3f posn=[%6.3f %6.3f %6.3f] vel=[%6.3f %6.3f %6.3f]\n",
		 id, mass, posn[0], posn[1], posn[2], vel[0], vel[1], vel[2]);
  printf("\tsprings to masspts: ");
  for (SLIST *s=slist; s; s=s->next) {
	int ids = s->spring->m1->id;
	if (ids==id)
	  ids = s->spring->m2->id;
	printf("%-4d ",ids);
  }
  printf("\n");
}

///////////////////////////////////////////////////////////
//    PROC: SPRING::SPRING(MASS *m1, MASS *m2, float kp, float kd, float len)
//    DOES: constructor for a SPRING with initialization
///////////////////////////////////////////////////////////

SPRING::SPRING(MASS *mass1, MASS *mass2, 
			   float spring_kp, float spring_kd, 
			   float spring_len, bool fVisible)
{
  m1 = mass1;
  SLIST *s1 = new SLIST;
  s1->spring = this;
  s1->next = mass1->slist;
  mass1->slist = s1;

  m2 = mass2;
  SLIST *s2 = new SLIST;
  s2->spring = this;
  s2->next = mass2->slist;
  mass2->slist = s2;

  kp = spring_kp;
  kd = spring_kd;
  len = spring_len;
  next = (SPRING *) 0;

  fVis = fVisible;
}

///////////////////////////////////////////////////////////
//    PROC: SPRING::print()
//    DOES: prints all info about a SPRING point
///////////////////////////////////////////////////////////

void SPRING::print()
{
  printf("m1=%-4d m2=%-4d kp=%6.3f kd=%6.3f len=%6.3f\n",
		 m1->id, m2->id, kp, kd, len);
}

///////////////////////////////////////////////////////////
//    PROC: NETWORK::add_mass(argc, argv)
//    DOES: adds a mass to a mass-spring network
///////////////////////////////////////////////////////////

void NETWORK::add_mass(int argc, char **argv)
{
  int nargs;
  STR args[6];

  if (argc<4 || argc>6) {
	printf("masspt usage:\n\t masspt <id_num> <mass> <x,y,z> [attr] [vx,vy,vz]\n");
  }

  FVEC posn;
  parse(argv[3],",", &nargs, args);
  posn[0] = atof(args[0]);
  posn[1] = atof(args[1]);
  posn[2] = atof(args[2]);

  float attr=0;
  if (argc>4)
	  attr = atof(argv[4]);

  printf("attr == %f\n",attr);

  FVEC vel;
  vel[0] = vel[1] = vel[2] = 0.0;
  if (argc==6) {
	parse(argv[5],",", &nargs, args);
	vel[0] = atof(args[0]);
	vel[1] = atof(args[1]);
	vel[2] = atof(args[2]);
  }

  MASS *m = new MASS( atoi(argv[1]),
					  atof(argv[2]),
					  posn,
					  vel,attr);
  m->next = masses;
  masses = m;
}

///////////////////////////////////////////////////////////
//    PROC: NETWORK::find_mass(int id)
//    DOES: finds a given mass point, given id
///////////////////////////////////////////////////////////

MASS *NETWORK::find_mass(int id)
{
  for (MASS *m=masses; m; m=m->next) {
	if (m->id == id)
	  return(m);
  }
  return 0;
}

///////////////////////////////////////////////////////////
//    PROC: NETWORK::add_spring(argc, argv)
//    DOES: adds a spring-damper unit
///////////////////////////////////////////////////////////

void NETWORK::add_spring(int argc, char **argv)
{
  int nargs;
  STR args[6];

  if (argc!=6 && argc!=7) {
	printf("spring usage:\n\t spring <mass1_id> <mass2_id> <kp> <kd> <len> [<visible?>]\n");
  }
  MASS *m1 = find_mass(atoi(argv[1]));
  MASS *m2 = find_mass(atoi(argv[2]));
  if (!m1 || !m2) {
	printf("Non-existing masspt\n");
	return;
  }
  
  bool fVisible = true;
  if( argc==7 )
	  switch( *argv[6] ) {
		  case 'n':
		  case 'N':
			  fVisible = false;
	  }

  SPRING *s = new SPRING( m1, m2, atof(argv[3]),
						  atof(argv[4]), atof(argv[5]), fVisible);
  s->next = springs;         // insert in linked list
  springs = s;
}

///////////////////////////////////////////////////////////
//    PROC: NETWORK::print()
//    DOES: prints contents of mass-spring network
///////////////////////////////////////////////////////////

void NETWORK::print()
{
  printf("MASSES:\n");
  for (MASS *m=masses; m; m=m->next) {
	m->print();
  }
  printf("SPRINGS:\n");
  for (SPRING *s=springs; s; s=s->next) {
	s->print();
  }
}

///////////////////////////////////////////////////////////
//    PROC: NETWORK::fix()
//    DOES: fixes the location of a masspt
///////////////////////////////////////////////////////////

void NETWORK::fix(int id)
{
  MASS *m = find_mass(id);
  if (!m) {
	printf("Mass %d not found\n",id);
	return;
  }
  m->type = FIXED;
}

///////////////////////////////////////////////////////////
//    PROC: NETWORK::draw()
//    DOES: displays mass-spring network
///////////////////////////////////////////////////////////

void NETWORK::draw()
{
  glColor3f(0.4,0,0.4);
  for (MASS *m=masses; m; m=m->next) {
	glPushMatrix();
	glTranslatef(m->posn[0], m->posn[1], m->posn[2]);
	glScalef(0.2,0.2,0.2);
	if (m->type==FIXED) {
	  glColor3f(0.7,0,0);
	  draw_tetra();
	  glColor3f(0.4,0,0.4);
	} else 
	  draw_tetra();
	glPopMatrix();
  }
  glColor3f(0,0,0);
  glBegin(GL_LINES);
  for (SPRING *s=springs; s; s=s->next) {
	if( !s->fVis )
		continue;
	glVertex3fv(s->m1->posn);
	glVertex3fv(s->m2->posn);
  }
  glEnd();

  glColor3f(0.8,0.8,0.8);    // draw a ground plane
  glBegin(GL_POLYGON);
  glVertex3fv(grnd[0]);
  glVertex3fv(grnd[1]);
  glVertex3fv(grnd[2]);
  glVertex3fv(grnd[3]);
//  glVertex3fv(-5,5,0);
  glEnd();

  glPushMatrix();
  glMultMatrixf((float *) sm);         // use shadow matrix
  glColor3f(0.6, 0.6, 0.6);  // a bit lighter shadow colour

  for (MASS *m=masses; m; m=m->next) {
	glPushMatrix();
	glTranslatef(m->posn[0], m->posn[1], m->posn[2]);
	glScalef(0.2,0.2,0.2);
	draw_tetra();
	glPopMatrix();
  }
  glBegin(GL_LINES);

  for (SPRING *s=springs; s; s=s->next) {
	glVertex3fv(s->m1->posn);
	glVertex3fv(s->m2->posn);
  }
  glEnd();

  glPopMatrix();
}

///////////////////////////////////////////////////////////
//    PROC: void draw_tetra()
//    DOES: draws a tetrahedron
///////////////////////////////////////////////////////////

void draw_tetra()
{
  static float v[6][3] = {
   -0.40,  0.5, -0.4,
	0.47,  0.0, -0.4,
   -0.40, -0.5, -0.4,
    0.00,  0.0,  0.47 
  };
	glBegin(GL_TRIANGLE_STRIP);
	glVertex3fv(v[0]);
	glVertex3fv(v[1]);
	glVertex3fv(v[2]);
	glVertex3fv(v[3]);
	glVertex3fv(v[1]);
	glVertex3fv(v[0]);
	glEnd();
}

///////////////////////////////////////////////////////////
//    PROC: net()
//    DOES: prints current mass-spring network
///////////////////////////////////////////////////////////

int net(ClientData clientData,
		   Tcl_Interp *interp,
		   int argc, char **argv)
{
  Network.print();
  return TCL_OK;
}

///////////////////////////////////////////////////////////
//    PROC: masspt()
//    DOES: creates a new mass point
///////////////////////////////////////////////////////////

int masspt(ClientData clientData,
		   Tcl_Interp *interp,
		   int argc, char **argv)
{
  Network.add_mass(argc, argv);
  return TCL_OK;
}

///////////////////////////////////////////////////////////
//    PROC: spring()
//    DOES: creates a new spring
///////////////////////////////////////////////////////////

int spring(ClientData clientData,
		   Tcl_Interp *interp,
		   int argc, char **argv)
{
  Network.add_spring(argc,argv);
  return TCL_OK;
}

///////////////////////////////////////////////////////////
//    PROC: fix()
//    DOES: fixes the location of a masspt
///////////////////////////////////////////////////////////

int fix(ClientData clientData,
		   Tcl_Interp *interp,
		   int argc, char **argv)
{
  if (argc!=2) {
	printf("Usage: fix <mass_id>\n");
	return TCL_OK;
  }
  Network.fix(atoi(argv[1]));
  return TCL_OK;
}



///////////////////////////////////////////////////////////
//    PROC: make_sm()
//    DOES: makes the shadow matrix
//          based on infinite source model, pg520 OpenGL Prog Guide
///////////////////////////////////////////////////////////

void make_sm() {

  // pick some values for dLight
  dLight[0] = 0.5;
  dLight[1] = -0.4;
  dLight[2] = -1;

  // calculate plane
  float x21, x31, y21, y31, z21, z31, x1, y1, z1;
  x21 = grnd[1][0] - grnd[0][0];
  x31 = grnd[2][0] - grnd[0][0];
  y21 = grnd[1][1] - grnd[0][1];
  y31 = grnd[2][1] - grnd[0][1];
  z21 = grnd[1][2] - grnd[0][2];
  z31 = grnd[2][2] - grnd[0][2];
  x1 = grnd[0][0];
  y1 = grnd[0][1];
  z1 = grnd[0][2];
  plane[0] = y21*z31 - z21*y31;			// A
  plane[1] = z21*x31 - x21*z31;			// B
  plane[2] = x21*y31 - y21*x31;			// C
  plane[3] = -x1 * plane[0] - y1 * plane[1] - z1 * plane[2];	// D

  printf("The ground plane: \n");
  prfvec(plane);
  
  idenmat( sm );
  sm[0][0] = plane[1]*dLight[1] + plane[2]*dLight[2];
  sm[0][1] = -plane[0]*dLight[1];
  sm[0][2] = -plane[0]*dLight[2];
  sm[1][0] = -plane[1]*dLight[0];
  sm[1][1] = plane[0]*dLight[0] + plane[2]*dLight[2];
  sm[1][2] = -plane[1]*dLight[2];
  sm[2][0] = -plane[2]*dLight[0];
  sm[2][1] = -plane[2]*dLight[1];
  sm[2][2] = plane[0]*dLight[0] + plane[1]*dLight[1];
  sm[3][0] = -plane[3]*dLight[0];
  sm[3][1] = -plane[3]*dLight[1];
  sm[3][2] = -plane[3]*dLight[2];
  sm[3][3] = plane[0]*dLight[0] + plane[1]*dLight[1]+plane[2]*dLight[2];

}



