///////////////////////////////////////////////
//    FILE: defsim1.C
//          more procedures and class functions for 
//          simulating mass-spring networks
///////////////////////////////////////////////

#include "tcl.h"
#include "defsim.h"
#include "lib/lib_defs.h"
#include "lib/lib_math.h"
#include "user.h"
#include <stdlib.h>
#include <GL/gl.h>

#include <assert.h>

extern NETWORK Network;

bool fClothModel = false;
bool fRecordOn = false;
float dragB = 0;
float dragN = 0;
float ALPHA = 0.001;	/* bounce dampening coefficient */
float FRICTION_K=0.8;	/* kinetic friction coefficient */

////// might get to this..., not sure if it would be useful though
//float FRICTION_S=1.0;	/* static friction coefficient */


void scalarmult( FVEC a, float f ) {
	a[0] *= f;
	a[1] *= f;
	a[2] *= f;
}

float dotproduct( FVEC a, FVEC b ) {
	return a[0]*b[0] + a[1]*b[1] + a[2]*b[2];
}

//
// debugging function, checks whether the float has gone NaN
//
bool isNaN( float f ) {
	char sz[128];
	sprintf(sz,"%f",f);
	if( !strcmp( "NaN", sz ) )
		return true;
	return false;
}

FVEC Fwind;
FVEC grnd[4];
FVEC grndN;
FVEC grndCenter;

class stack {
	struct element {
		void* data;
		element* next;

		element(): data(NULL), next(NULL) {}
	};

public:
	stack(): top(NULL) {}

	void push( void* p ) {
		element* pel = new element;
		if( pel ) {
			pel->data = p;
			pel->next = top;
			top = pel;
		}
	}

	void* pop() {
		if( !top )
			return NULL;

		element* pel = top;
		if( !pel->next ) {
			top = NULL;
			return pel->data;
		}

		while( pel->next->next )
			pel = pel->next;

		element* p = pel->next;
		pel->next = NULL;
		return p->data;
	}

	bool empty() { return (top)? false: true; }

private:
	element* top;
};

///////////////////////////////////////////////////////////
//    PROC: NETWORK::sim()
//    DOES: computes accelerations for masspts
///////////////////////////////////////////////////////////

void NETWORK::sim()
{
  MASS *m;

  for (m=masses; m; m=m->next) {

	  // reset the force felt to just the gravity force
	  m->force[0] = m->force[1] = 0;
//	  m->force[2] = -9.8 * m->mass;
	  m->force[2] = 0;

	  // now add the wind force
	  vecadd( m->force, m->force, Fwind );

	  // and air drag based on velocity from previous calculation cycle
	  if( dragB && vecsize(m->vel) ) {
		  FVEC f;
		  veccp(f,m->vel);
		  scalarmult(f,-1);
		  norm(f);
		  scalarmult(f, dragB * pow( vecsize(m->vel), dragN ) );

		  vecadd( m->force, m->force, f );
	  }

	  for( MASS* mAttr=masses; mAttr; mAttr=mAttr->next ) {
		  if( mAttr->attraction && mAttr!=m ) {
			  FVEC f;
			  vecsub(f,mAttr->posn,m->posn);
			  float dist=vecsize(f);
			  norm(f);
			  scalarmult(f,mAttr->attraction * m->mass / dist / dist );

			  vecadd( m->force, m->force, f );
		  }
	  }

	  for (SPRING *s=springs; s; s=s->next) {	// go through all springs
		if( s->m1==m || s->m2==m ) {			// if this spring is attached to mass
			FVEC ab;			// vector from this mass to the other mass
			FVEC old_ab;		// previous iteration's 'ab' vector
			if( s->m1==m ) {
				vecsub( ab, s->m2->posn, s->m1->posn );
				vecsub( old_ab, s->m2->prevposn, s->m1->prevposn );
			}
			else {
				vecsub( ab, s->m1->posn, s->m2->posn );
				vecsub( old_ab, s->m1->prevposn, s->m2->prevposn );
			}

			// the scalar "dotted l"
			float ldot = (vecsize(ab) - vecsize(old_ab)) / DTsim;	

			FVEC Fv;			// vector form of the total force
			float Fmag = s->kp * ( vecsize(ab) - s->len) + s->kd * ldot;
			float Fkp = s->kp * ( vecsize(ab) - s->len);
			float Fkd = s->kd * ldot;
			veccp( Fv, ab );
			norm( Fv );
			scalarmult( Fv, Fmag );

			// now add this force to total force felt
			vecadd( m->force, m->force, Fv );
	  }
	}
  }

// collision detection, and if so, add the frictional forces; also
// calculate the acceleration
for( m=masses; m; m=m->next ) {

	// do collision-detection
/* *** this method loses precision the further you are from the origin
	FVEC test;
	vecsub( test, m->posn, grndCenter );
	if( dotproduct(test,grndN) <= 0 ) {		// got a collision
*/

	// here's a better way to test for collision???
	// just plug in m->posn into the ground plane equation
	float overunder = plane[0]*m->posn[0] + plane[1]*m->posn[1] +
					  plane[2]*m->posn[2] + plane[3];
	if( overunder >= 0 ) {
		// figure out where path intersected
		float a,b;
		a = dist2ground( m->posn );
		b = dist2ground( m->prevposn );

		if( a+b ) { // only correct position and velocity if need to

			FVEC correction;
			vecsub( correction, m->prevposn, m->posn );
			scalarmult( correction, a/(a+b) );
			vecadd( m->posn, m->posn, correction );

			if( dotproduct(m->vel,grndN) > 0 ) {	// already bouncing back
				printf("Strange: already bouncing back...\n");
			}
			else {
				// modify the velocity components
				FVEC vn, vt;		// the normal,tangent parts of velocity
				veccp( vn, grndN );
				scalarmult( vn, dotproduct( m->vel, grndN ) );
				vecsub( vt, m->vel, vn );

				scalarmult( vn, - ALPHA );
				vecadd( m->vel, vn, vt );
			}
		}

		// sliding friction force calculation
		FVEC Ft, Fn, Vn, Vt, Forigt;
		veccp( Fn, grndN );
		scalarmult( Fn, dotproduct( m->force, grndN ) );
		vecsub( Forigt, m->force, Fn );
		veccp( Vn, grndN );
		scalarmult( Vn, dotproduct( m->vel, grndN ) );
		vecsub( Vt, m->vel, Vn );

		if( vecsize(Vt) ) {
			veccp( Ft, Vt );
			norm( Ft );
			scalarmult( Ft, - FRICTION_K * vecsize(Fn) );

			vecadd( m->force, m->force, Ft );

			// oh yeah, add the normal force exerted by the ground
			scalarmult( Fn, -1 );		// now its the ground pushing
			vecadd( m->force, m->force, Fn );
		}
	}

	veccp( m->acc, m->force );
	scalarmult( m->acc, 1/m->mass );
  }
}

///////////////////////////////////////////////////////////
//    PROC: NETWORK::integ()
//    DOES: integrates the system state
///////////////////////////////////////////////////////////

const float ELASTICITY_MAX = 1.20;

void NETWORK::integ()
{
  for (MASS *m=masses; m; m=m->next) {
	if (m->type==FIXED) continue;
	veccp( m->prevposn, m->posn );	// this position now becomes the previous
    veccp( m->prevvel, m->vel );    // ditto for velocity
	for (int n=0; n<3; n++) {
	  m->posn[n] += m->vel[n]*DTsim + 0.5*m->acc[n]*DTsim*DTsim;
	  m->vel[n] += m->acc[n]*DTsim;
	}

  }

	// now to do the stretch-limiting
	if( fClothModel ) {
		stack stk;

		for( MASS* m=masses; m; m=m->next ) {
			m->Flag = false;
			if( m->type == FIXED ) {
				stk.push( m );
			}
		}
		if( stk.empty() )
			stk.push( masses );			// start with something

		while( !stk.empty() ) {
			MASS* m = (MASS*)stk.pop();
			if( m->Flag )
				continue;

			for( SPRING *s=springs; s; s=s->next ) {

				if( s->m1->Flag || s->m2->Flag )
					continue;				// this spring was already done

				if( s->m1 == m || s->m2 == m ) {

					stk.push( s->m2 );
					stk.push( s->m1 );

					FVEC ab;
					vecsub( ab, s->m1->posn, s->m2->posn );
					if( vecsize( ab ) > ELASTICITY_MAX * s->len ) {
						// this spring is stretched too much; contract
						// case 1: both ends free; contract equally from both ends
						if( (s->m1->type == FREE) && (s->m2->type == FREE) ) {
							FVEC v;
							FVEC margin;
							vecsub( v, s->m2->posn, s->m1->posn );
							float f = (s->len * ELASTICITY_MAX) / vecsize( v );

							veccp( margin, v );

							scalarmult( v, f );

							vecsub( margin, margin, v );
							scalarmult( margin, 0.5 );
							
							FVEC t;
							vecadd( t, s->m1->posn, margin );
							s->m1->inverseDynamics( t );
							vecadd( t, s->m1->posn, v );
							s->m2->inverseDynamics( t );
						}
						// case 2: end at mass 1 fixed
						else if( s->m1->type == FIXED && s->m2->type == FREE ) {
							FVEC v;
							vecsub( v, s->m2->posn, s->m1->posn );
							float vs = vecsize( v );
							float f = (s->len * ELASTICITY_MAX) / vs;
							scalarmult( v, f );

							FVEC t;
							vecadd( t, s->m1->posn, v );
							s->m2->inverseDynamics( t );
						}
						// case 3: end at mass 2 fixed
						else if( s->m2->type == FIXED && s->m1->type == FREE ) {
							FVEC v;
							vecsub( v, s->m1->posn, s->m2->posn );
							float f = (s->len * ELASTICITY_MAX) / vecsize( v );
							scalarmult( v, f );

							FVEC t;
							vecadd( t, s->m2->posn, v );
							s->m1->inverseDynamics( t );
						}
						// case 4: both ends fixed; can't do much at all
					}
				}
			}

			m->Flag = true;
		}
	}
}

///////////////////////////////////////////////////////////
//    PROC: wind()
//    DOES: sets the Fwind vector
///////////////////////////////////////////////////////////

int wind(ClientData clientData,
		   Tcl_Interp *interp,
		   int argc, char **argv)
{
  if( argc!=4 ) {
	  printf("wind usage:\n\t wind <x> <y> <z>\n");
  }
  
  for( int i=0; i<3; i++ )
	  Fwind[i] = atof(argv[1+i]);

  return TCL_OK;
}

///////////////////////////////////////////////////////////
//    PROC: ground()
//    DOES: sets the four vertices of the ground plane
//          based on plane centerpoint and its normal
///////////////////////////////////////////////////////////

const float GROUND_EDGE=10.0;

int ground(ClientData clientData,
		   Tcl_Interp *interp,
		   int argc, char **argv)
{
  if( argc!=5 ) {
	  printf("ground usage:\n\t ground <n.x> <n.y> <n.z> <height@0,0>\n");
  }
  
  for( int i=0; i<3; i++ ) {
	  grndN[i] = atof(argv[1+i]);
	  grndCenter[i] = 0;
  }
  grndCenter[2] = atof(argv[4]);

  norm( grndN );

  FVEC t,u,v;
  for( int i=0; i<3; i++ )
	  t[i] = 1234.5678;
  xprodnorm( u, grndN, t );
  scalarmult( u, GROUND_EDGE/2 );
  xprodnorm( v, grndN, u );
  scalarmult( v, GROUND_EDGE/2 );

  vecadd( grnd[0], grndCenter, u );
  vecadd( grnd[0], grnd[0], v );
  vecadd( grnd[1], grndCenter, u );
  vecsub( grnd[1], grnd[1], v );
  vecsub( grnd[2], grndCenter, u );
  vecsub( grnd[2], grnd[2], v );
  vecsub( grnd[3], grndCenter, u );
  vecadd( grnd[3], grnd[3], v );


  // make the shadow matrix once and for all (the rest of this run)
  make_sm();

  return TCL_OK;
}

///////////////////////////////////////////////////////////
//    PROC: dist2ground( FVEC )
//    DOES: returns the distance of point to ground plane
///////////////////////////////////////////////////////////
float dist2ground( FVEC v ) {
	if( !plane[0] && !plane[1] && !plane[2] ) {
		printf("Invalid plane given\n");
		return 0;
	}
	return fabs((plane[0]*v[0] + plane[1]*v[1] + plane[2]*v[2] + plane[3]) /
			sqrt(plane[0]*plane[0]+plane[1]*plane[1]+plane[2]*plane[2]));
}

///////////////////////////////////////////////////////////
//    PROC: cloth_model(...)
//    DOES: turns on spring elasticity constraints
///////////////////////////////////////////////////////////
int cloth_model(ClientData clientData, Tcl_Interp *interp,
		   int argc, char **argv) {

	fClothModel = true;
	return TCL_OK;
}


///////////////////////////////////////////////////////////
//    PROC: MASS::inverseDynamics()
//    DOES: recalculates/corrects acceleration and velocity
//			of the mass based upon the new particle position
//		    specified by the user
///////////////////////////////////////////////////////////

void MASS::inverseDynamics( FVEC newpos ) {
	
	// now figure out the corrected acceleration
	FVEC u,v;
	vecsub( u, newpos, prevposn );
	veccp( v, prevvel );
	scalarmult( v, DTsim );
	vecsub( u, u, v );
	scalarmult( u, 2/DTsim/DTsim );
	veccp( acc, u );

	// next, the corrected velocity
	scalarmult( u, DTsim );
	vecadd( vel, prevvel, u );

	// finally, just set the particle position to the desired position
	veccp( posn, newpos );
}

///////////////////////////////////////////////////////////
//    PROC: dt_disp(...)
//    DOES: changes DTdisp
///////////////////////////////////////////////////////////
int dt_disp(ClientData clientData, Tcl_Interp *interp,
		   int argc, char **argv) {
 
	  if( argc!=2 ) {
		  printf("dt_disp usage:\n\t dt_disp <time>\n");
		  return TCL_OK;
	  }

	  DTdisp = atof(argv[1]);

	return TCL_OK;
}

///////////////////////////////////////////////////////////
//    PROC: dt_sim(...)
//    DOES: changes DTsim
///////////////////////////////////////////////////////////
int dt_sim(ClientData clientData, Tcl_Interp *interp,
		   int argc, char **argv) {
 
	  if( argc!=2 ) {
		  printf("dt_sim usage:\n\t dt_sim <time>\n");
		  return TCL_OK;
	  }

	  DTsim = atof(argv[1]);

	return TCL_OK;
}
///////////////////////////////////////////////////////////
//    PROC: record_on()
//    DOES: turns on recording of frames
///////////////////////////////////////////////////////////
int record_on(ClientData clientData, Tcl_Interp *interp,
		   int argc, char **argv) {
 
	fRecordOn = true;
	return TCL_OK;
}

///////////////////////////////////////////////////////////
//    PROC: drag_coeffs(...)
//    DOES: sets the air drag coefficients
///////////////////////////////////////////////////////////
int drag_coeffs(ClientData clientData, Tcl_Interp *interp,
		   int argc, char **argv) {
 
	  if( argc!=3 ) {
		  printf("drag_coeffs:\n\t drag_coeffs <b><n>\n");
		  // where F_drag = (b)(vel^n)
		  return TCL_OK;
	  }

	  dragB = atof(argv[1]);
	  dragN = atof(argv[2]);

	return TCL_OK;
}

///////////////////////////////////////////////////////////
//    PROC: bounce_alpha(...)
//    DOES: sets the percentage of energy kept in a bounce
///////////////////////////////////////////////////////////
int bounce_alpha(ClientData clientData, Tcl_Interp *interp,
		   int argc, char **argv) {
 
	  if( argc!=2 ) {
		  printf("bounce_alpha:\n\t bounce_alpha <alpha>\n");
		  return TCL_OK;	// want to keep reading the file properly
	  }

	  ALPHA = atof(argv[1]);

	return TCL_OK;
}

///////////////////////////////////////////////////////////
//    PROC: friction_kcoeff(...)
//    DOES: sets the kinetic friction coefficient
///////////////////////////////////////////////////////////
int friction_kcoeff(ClientData clientData, Tcl_Interp *interp,
		   int argc, char **argv) {
 
	  if( argc!=2 ) {
		  printf("friction_kcoeff:\n\t friction_kcoeff <coeff>\n");
		  return TCL_OK;	// want to keep reading the file properly
	  }

	  FRICTION_K = atof(argv[1]);

	return TCL_OK;
}

