/************************************************************
	FILE:		lib_image.c
	CONTAINS:	functions for reading and writing SGI RGB-format images
************************************************************/

#include <stdlib.h>
#include "tcl.h"
#include "glaux.h"

#include "lib_defs.h"
#include "lib_globals.h"
#include "lib_com.h"
#include "lib_image.h"
#include "lib_draw.h"

IMAGE *Image;

/*********************************************************
    PROC:  mymalloc()
    DOES:  allocates memory
**********************************************************/

char *mymalloc(int size)
{
  char *buf;

  buf = (char *) malloc(size);
  if (!buf) {
    fprintf(stderr,"malloc() failure\n");
    exit(-1);
  }
  return(buf);
}

/*********************************************************
    PROC:  prheader(h)
    DOES:  prints the SGI file header
**********************************************************/

prheader(SGIHEADER *h)
{
  printf("%5d Magic file # (474)\n",h->magic);
  fflush(stdout);
  printf("%5d Storage format (0=non-rle, 1=rle)\n",(int) h->storage);
  printf("%5d Number of bytes per pixel component (probably 1)\n",h->bpc);
  printf("%5u Dimension (2)\n",h->dim);
  printf("%5u Xsize\n",h->xsize);
  printf("%5u Ysize\n",h->ysize);
  printf("%5u Zsize (3=24-bit RGB, 4=RGBA)\n",h->zsize);
  printf("%5u Pixmin (probably 0)\n",h->pixmin);
  printf("%5u Pixmax (probably 255)\n",h->pixmax);
  printf("'%s'  Image name\n",h->iname);
  printf("%ld Colormap (0=normal, 1=dithered, 2=mapped, 3=cmap-data)\n",h->cmap);
}

/*********************************************************
    PROC:  getbyte()
    DOES: 
**********************************************************/

unsigned char getbyte(FILE *inf)
{
  unsigned char buf;

  fread(&buf,1,1,inf);
  return(buf);
}

/*********************************************************
    PROC:  getshort()
    DOES: 
**********************************************************/

unsigned short getshort(FILE *inf) 
{ 
  unsigned char buf[2]; 
 
  fread(buf,2,1,inf); 
  return (buf[0]<<8)+(buf[1]<<0); 
} 

/*********************************************************
    PROC:  getlong()
    DOES: 
**********************************************************/

long getlong(FILE *inf) 
{ 
  unsigned char buf[4]; 
 
  fread(buf,4,1,inf); 
  return (buf[0]<<24)+(buf[1]<<16)+(buf[2]<<8)+(buf[3]<<0); 
} 

/*********************************************************
    PROC:  putbyte()
    DOES: 
**********************************************************/

putbyte(FILE *outf, unsigned char val) 
{ 
    unsigned char buf[1]; 

    buf[0] = val; 
        fwrite(buf,1,1,outf); 
    } 
 
/*********************************************************
    PROC:  putshort()
    DOES: 
**********************************************************/

putshort(FILE *outf, unsigned short val) 
{ 
        unsigned char buf[2]; 
 
        buf[0] = (val>>8); 
        buf[1] = (val>>0); 
        fwrite(buf,2,1,outf); 
} 
 
/*********************************************************
    PROC:  putlong()
    DOES:  
**********************************************************/

static int putlong(FILE *outf, unsigned long val) 
{ 
	unsigned char buf[4]; 
 
	buf[0] = (val>>24); 
	buf[1] = (val>>16); 
	buf[2] = (val>>8); 
	buf[3] = (val>>0); 
	return fwrite(buf,4,1,outf); 
} 
 
/*********************************************************
    PROC:  find_image()
    DOES:  finds image of the given name
**********************************************************/

IMAGE *find_image(char *name)
{
  IMAGE *image;

  for (image=Image; image; image=image->next) {
    if (!strcmp(name,image->name))
      return(image);
  }
  return((IMAGE *) NULL);
}

/*********************************************************
    PROC:  newimage()
    DOES:  creates a new image
**********************************************************/

IMAGE *newimage(char *name)
{
  IMAGE *i;

  i = (IMAGE *) mymalloc(sizeof(IMAGE));
  strcpy(i->name,name);
  i->data = (char *) NULL;
  i->next = Image;
  Image = i;
  return(i);
}

/*********************************************************
    PROC:  texload
    DOES:  reads in an rgb image
**********************************************************/

int texload(ClientData clientData,
		   Tcl_Interp *interp,
		   int argc, char **argv)
{ 
  FILE *inf; 
  IMAGE *image;
  SGIHEADER *h;

  if (argc<3) {
	cerr(argc,argv,"Bad # args");
	return TCL_OK;
  }
              /* open file */
  inf = fopen(argv[2],"r"); 
  if(!inf) { 
      fprintf(stderr,"sgiimage: can't open input file\n"); 
	  return TCL_OK;
  } 
  image = find_image(argv[1]);        /* check for existing image */
  if (!image) {
    image = newimage(argv[1]);
  }
  h = &(image->h);
  read_header(h,inf);          /* read header record */
  strcpy(image->name,argv[1]);
  if (DEbug)
    prheader(h);
                                       /* allocate image space */
  if (image->data)
    free(image->data);
  image->data = mymalloc(h->bpc*h->zsize*h->xsize*h->ysize);
  image->data2 = mymalloc(h->bpc*h->zsize*h->xsize*h->ysize);
  switch(h->storage) {
    case VERBATIM:
      read_verbatim(inf,image->data,h);
      break;
    case RLE:
      read_rle(inf,image->data,h);
      break;
    default:
      fprintf(stderr,"File storage format unrecognized\n");
      exit(-1);
  }
  image_convert(image);
  fclose(inf); 
  TExtures = TRUE;
  tex_init();
  return TCL_OK;
}

/************************************************************
	PROC:	image_convert()
	DOES:	converts from row-column-byte to column-row-byte
************************************************************/

image_convert(IMAGE *i)
{
  int x,y,z;

  for (x=0; x<i->h.xsize; x++) {
    for (y=0; y<i->h.ysize; y++) {
      for (z=0; z<3; z++) {
	i->data2[x*i->h.ysize*3 + y*3 + z] = i->data[y*i->h.xsize*3 + x*3 + z];
      }
    }
  }
}

/************************************************************
	PROC:	print()
	DOES:	prints the contents of an image
************************************************************/

void print(int nargs, STR args[])
{
    int row;
    IMAGE *image;
    SGIHEADER *h;

    if (nargs<2 || nargs>3) {
      printf("Bad number of arguments\n");
      return;
    }
    image = find_image(args[1]);
    if (!image) {
      printf("Image '%s' not found\n",args[1]);
      return;
    }
    h = &(image->h);
    if (nargs==2) 
	pr_image(image->data,h);
    else {
	row = atoi(args[2]);
	if (row<0 || row>=h->ysize) {
	    printf("Error: the image only has %d scanlines\n",h->ysize);
	    return;
	}
	prscanline(row,image->data,h);
    }
}

/************************************************************
	PROC:	save()
	DOES:	saves the image to a file
************************************************************/

void save(int nargs, STR args[])
{
  IMAGE *image;

  image = find_image(args[1]);
  if (!image) {
    printf("Image '%s' not found\n",args[1]);
    return;
  }
  put_image_verbatim(image->data,&(image->h),args[2]);
}

/*********************************************************
    PROC:  put_image_verbatim()
    DOES:  writes verbatim image in SGI format
**********************************************************/

put_image_verbatim(char *image, SGIHEADER *h, char *name)
{
        int i, x, y; 
	int offset;
	int c;
	FILE *of;
	char *ptr;

        of = fopen(name,"w");
	if (!of) {
	  fprintf(stderr,"Unable to open '%s' for write.\n",name);
	  exit(-1);
	}
        putshort(of,474);       /* MAGIC 	       	*/ 
        putbyte(of,0);          /* STORAGE is VERBATIM 	*/ 
        putbyte(of,h->bpc);     /* BPC             	*/ 
        putshort(of,h->dim);    /* DIMENSION */
        putshort(of,h->xsize);  /* XSIZE               	*/ 
        putshort(of,h->ysize);  /* YSIZE               	*/ 
        putshort(of,h->zsize);  /* ZSIZE               	*/ 
        putlong(of,0);          /* PIXMIN is 0         	*/ 
        putlong(of,255);        /* PIXMAX is 255       	*/ 
        for(i=0; i<4; i++)      /* DUMMY 4 bytes 	*/ 
            putbyte(of,0); 
        fwrite(h->iname,80,1,of);  /* IMAGENAME  		*/ 
        putlong(of,0);          /* COLORMAP is 0 	*/ 
        for(i=0; i<404; i++)    /* DUMMY 404 bytes 	*/ 
            putbyte(of,0); 
	for (c=0; c<h->zsize; c++) {
          for(y=0; y<h->ysize; y++) {   /* image */
	    ptr = image + y*h->xsize*h->zsize + c;
            for (x=0; x<h->xsize; x++) {
		putbyte(of,*ptr);
		ptr += h->zsize;
       	    }
          } 
        }
        fclose(of); 
}

/*********************************************************
    PROC:  pr_image()
    DOES:  prints image data 
**********************************************************/

pr_image(char *image, SGIHEADER *h)
{
  int y;

  for (y=0; y<h->ysize; y++) {
    prscanline(y,image,h);
  }
}

/************************************************************
	PROC:	prscanline()
	DOES:	dumps the contents of a scanline
************************************************************/

prscanline(int y, char *image, SGIHEADER *h)
{
  int x;
  int r,g,b,a;
   
    image += y*h->xsize*h->zsize;
    printf("scanline %d\n\t",y);
    for (x=0; x<h->xsize; x++) {
	if (!(x%10)) {
	    printf("\n\t");
	}
        switch(h->zsize) {
        case 1:
  	  printf("%3d ",(unsigned char) *image);
	  image++;
	  break;
	case 3:
	  r = (unsigned char) *image;
	  g = (unsigned char) *(image+1);
	  b = (unsigned char) *(image+2);
	  printf("%3d %3d %3d ",r,g,b);
	  image += 3;
	  break;
	case 4:
	  r = (unsigned char) *image;
	  g = (unsigned char) *(image+1);
	  b = (unsigned char) *(image+2);
	  a = (unsigned char) *(image+3);
	  printf("%3d %3d %3d %3d ",r,g,b,a);
	  image += 4;
	  break;
      }
    }
    printf("\n");
}

/*********************************************************
    PROC:  read_verbatim()
    DOES:  reads verbatim image data
**********************************************************/

read_verbatim(FILE *inf, char *image, SGIHEADER *h)
{
  int c;
  int x,y;
  char *optr;

  printf("Reading verbatim file format\n");
  fseek(inf,512,SEEK_SET); 
  for (c=0; c<h->zsize; c++) {
    for (y=0; y<h->ysize; y++) {
      fread(Scanline,h->xsize,1,inf);
      optr = image + y*h->xsize*h->zsize + c;
      for (x=0; x<h->xsize; x++) {
	*optr = Scanline[x];
	optr += h->zsize;
      }
    }
  }
}

/*********************************************************
    PROC:  read_rle()
    DOES:  reads rle image data
**********************************************************/

read_rle(FILE *inf, char *image, SGIHEADER *h)
{
        int x, y; 
	int tablen;
	long *starttab;
	long *lengthtab;
	int rleoffset,rlelength;
	int c;

        printf("Reading rle file format\n");
	tablen = h->ysize*h->zsize*sizeof(long); 
	starttab = (long *)mymalloc(tablen); 
	lengthtab = (long *)mymalloc(tablen); 
	fseek(inf,512,SEEK_SET); 
        readlongtab(inf,starttab,h->ysize*h->zsize); 
	readlongtab(inf,lengthtab,h->ysize*h->zsize); 
	for (c=0; c<h->zsize; c++) {
	  for (y=0; y<h->ysize; y++) {
	    rleoffset = starttab[y + c*h->ysize];
	    rlelength = lengthtab[y + c*h->ysize];
/*
	    printf("rleoffset=%d  rlelength=%d\n",rleoffset,rlelength);
*/
	    getiline(inf,image + y*h->xsize*h->zsize + c, 
		     rleoffset, rlelength,h->zsize);
	  }
	} 
	free(starttab);
	free(lengthtab);
} 

/*********************************************************
    PROC:  read_header()
    DOES:  reads header name
**********************************************************/

read_header(SGIHEADER *h, FILE *inf)
{
  int i;

        h->magic = getshort(inf);       /* MAGIC 	       	*/ 
	h->storage = getbyte(inf);      /* STORAGE               */
        h->bpc = getbyte(inf);          /* BPC is 1 (bytes per pixel channel) */
	h->dim = getshort(inf);         /* DIMENSION is 2      	*/ 
        h->xsize = getshort(inf);       /* XSIZE               	*/ 
        h->ysize = getshort(inf);       /* YSIZE               	*/ 
        h->zsize = getshort(inf);       /* ZSIZE               	*/ 
        h->pixmin = getlong(inf);       /* PIXMIN value           */
        h->pixmax = getlong(inf);       /* PIXMAX is 255       	*/ 
        for(i=0; i<4; i++)            /* DUMMY 4 bytes 	*/ 
            getbyte(inf); 
	fread(h->iname,80,1,inf);       /* image name */
        h->cmap = getlong(inf);         /* colour map ID */
        getlong(inf);                /* COLORMAP is 0 	*/ 
/*
        for(i=0; i<404; i++)
            getbyte(inf); 
*/
}

/*********************************************************
    PROC:  getiline()
    DOES:  gets an image line
**********************************************************/

getiline(FILE *inf, char *image,
		 int rleoffset, int rlelength, int zsize)
{
  fseek(inf,rleoffset,SEEK_SET);     /* wind file to the right point */
  fread(Scanline,rlelength,1,inf);   /* read compressed line */
  expandrow((unsigned char *) image, 
			(unsigned char *) Scanline,zsize); /* expand the compressed line */
}

/*********************************************************
    PROC:  expandrow()
    DOES:  expand an rle-compressed row
**********************************************************/

void expandrow(unsigned char *optr,
		  unsigned char *iptr, int z) 
{ 
  unsigned char pixel, count; 
	     
  while(1) { 
    pixel = *iptr++; 
    if ( !(count = (pixel & 0x7f)) )   /* count=0, no image data */
      return; 
    if(pixel & 0x80) {                 /* straight copy of following bytes */
      while(count--) { 
	*optr = *iptr++; 
	  optr+=z; 
      } 
    } else {                           /* duplicate following byte */
      pixel = *iptr++; 
      while(count--) { 
	*optr = pixel; 
	optr+=z; 
      } 
    } 
  } 
} 
 
/*********************************************************
    PROC:  readlongtab()
    DOES:  fills a table
**********************************************************/

readlongtab(FILE *inf, long *table, int size)
{
  int i;

  for (i=0; i<size; i++) {
    table[i] = getlong(inf);
  }
}

/*********************************************************
    PROC:  texhead(nargs,args)
    DOES:  prints the contents of an image header
**********************************************************/

int texhead(ClientData clientData,
		   Tcl_Interp *interp,
		   int argc, char **argv)
{
  IMAGE *i;

  if (argc<2) {
	cerr(argc,argv,"Bad # of args");
	return TCL_OK;
  }
  i = find_image(argv[1]);
  if (!i) {
    printf("Image '%s' not found\n",argv[1]);
    return TCL_OK;
  }
  prheader(&(i->h));
  return TCL_OK;
}









